/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <cerrno>

#include <android-base/file.h>
#include <android-base/logging.h>

#include "Thermal.h"
#include "thermal-helper.h"

namespace android {
namespace hardware {
namespace thermal {
namespace V1_0 {
namespace implementation {

namespace {

using ::android::hardware::thermal::V1_0::ThermalStatus;
using ::android::hardware::thermal::V1_0::ThermalStatusCode;
using ::android::hardware::Void;

template <typename T, typename U>
Return<void> setFailureAndCallback(
    T _hidl_cb, hidl_vec<U> data, const std::string& debug_msg) {
    ThermalStatus status;
    status.code = ThermalStatusCode::FAILURE;
    status.debugMessage = debug_msg;
    _hidl_cb(status, data);
    return Void();
}

template <typename T, typename U>
Return<void> setInitFailureAndCallback(T _hidl_cb, hidl_vec<U> data) {
    return setFailureAndCallback(
        _hidl_cb, data, "Failure initializing thermal HAL");
}

}  // namespace

// Methods from ::android::hardware::thermal::V1_0::IThermal.
Return<void> Thermal::getTemperatures(getTemperatures_cb _hidl_cb) {
    ThermalStatus status;
    status.code = ThermalStatusCode::SUCCESS;
    hidl_vec<Temperature> temperatures;

    if (!thermal_helper.isInitializedOk()) {
        LOG(ERROR) << "ThermalHAL not initialized properly.";
        return setInitFailureAndCallback(_hidl_cb, temperatures);
    }

    if (!thermal_helper.fillTemperatures(&temperatures)) {
        return setFailureAndCallback(_hidl_cb, temperatures,
                "Failed to read thermal sensors.");
    }

    for (const auto& t : temperatures) {
        LOG(DEBUG) << "getTemperatures "
                   << " Type: " << android::hardware::thermal::V1_0::toString(t.type)
                   << " Name: " << t.name
                   << " CurrentValue: " << t.currentValue
                   << " ThrottlingThreshold: " << t.throttlingThreshold
                   << " ShutdownThreshold: " << t.shutdownThreshold
                   << " VrThrottlingThreshold: " << t.vrThrottlingThreshold;
    }
    _hidl_cb(status, temperatures);
    return Void();
}

Return<void> Thermal::getCpuUsages(getCpuUsages_cb _hidl_cb) {
    ThermalStatus status;
    status.code = ThermalStatusCode::SUCCESS;
    hidl_vec<CpuUsage> cpu_usages;

    if (!thermal_helper.isInitializedOk()) {
        return setInitFailureAndCallback(_hidl_cb, cpu_usages);
    }

    if (!thermal_helper.fillCpuUsages(&cpu_usages)) {
        return setFailureAndCallback(_hidl_cb, cpu_usages,
            "Failed to get CPU usages.");
    }

    for (const auto& usage : cpu_usages) {
        LOG(DEBUG) << "getCpuUsages "
                   << " Name: " << usage.name
                   << " Active: " << usage.active
                   << " Total: " << usage.total
                   << " IsOnline: " << usage.isOnline;
    }
    _hidl_cb(status, cpu_usages);
    return Void();
}

Return<void> Thermal::getCoolingDevices(getCoolingDevices_cb _hidl_cb) {
    ThermalStatus status;
    status.code = ThermalStatusCode::SUCCESS;
    hidl_vec<CoolingDevice> cooling_devices;

    if (!thermal_helper.isInitializedOk()) {
        return setInitFailureAndCallback(_hidl_cb, cooling_devices);
    }
    LOG(DEBUG) << "No Cooling Devices";
    _hidl_cb(status, cooling_devices);
    return Void();
}

Return<void> Thermal::debug(const hidl_handle& handle, const hidl_vec<hidl_string>&) {
    if (handle != nullptr && handle->numFds >= 1) {
        int fd = handle->data[0];
        std::ostringstream dump_buf;

        if (!thermal_helper.isInitializedOk()) {
            dump_buf << "ThermalHAL not initialized properly." << std::endl;
        } else {
            hidl_vec<Temperature> temperatures;
            hidl_vec<CpuUsage> cpu_usages;

            dump_buf << "getTemperatures:" << std::endl;
            if (!thermal_helper.fillTemperatures(&temperatures)) {
                dump_buf << "Failed to read thermal sensors." << std::endl;
            }

            for (const auto& t : temperatures) {
                dump_buf << "Name: " << t.name
                         << " Type: " << android::hardware::thermal::V1_0::toString(t.type)
                         << " CurrentValue: " << t.currentValue
                         << " ThrottlingThreshold: " << t.throttlingThreshold
                         << " ShutdownThreshold: " << t.shutdownThreshold
                         << " VrThrottlingThreshold: " << t.vrThrottlingThreshold
                         << std::endl;
            }

            dump_buf << "getCpuUsages:" << std::endl;
            if (!thermal_helper.fillCpuUsages(&cpu_usages)) {
                dump_buf << "Failed to get CPU usages." << std::endl;
            }

            for (const auto& usage : cpu_usages) {
                dump_buf << "Name: " << usage.name
                         << " Active: " << usage.active
                         << " Total: " << usage.total
                         << " IsOnline: " << usage.isOnline
                         << std::endl;
            }

        }
        std::string buf = dump_buf.str();
        if (!android::base::WriteStringToFd(buf, fd)) {
            PLOG(ERROR) << "Failed to dump state to fd";
        }
        fsync(fd);
    }
    return Void();
}

}  // namespace implementation
}  // namespace V1_0
}  // namespace thermal
}  // namespace hardware
}  // namespace android
